=== modified file 'bzrlib/repofmt/groupcompress_repo.py'
--- bzrlib/repofmt/groupcompress_repo.py	2009-09-07 06:01:48 +0000
+++ bzrlib/repofmt/groupcompress_repo.py	2009-09-07 07:41:17 +0000
@@ -601,6 +601,7 @@
         new_revisions_keys = key_deps.get_new_keys()
         no_fallback_inv_index = self.repo.inventories._index
         no_fallback_chk_bytes_index = self.repo.chk_bytes._index
+        no_fallback_texts_index = self.repo.texts._index
         inv_parent_map = no_fallback_inv_index.get_parent_map(
             new_revisions_keys)
         # Are any inventories for corresponding to the new revisions missing?
@@ -627,14 +628,16 @@
         def root_keys_for_inv(inv):
             return [inv.id_to_entry.key(),
                     inv.parent_id_basename_to_file_id.key()]
-        interesting = []
-        uninteresting = []
+        interesting_key_lists = ([], [])
+        uninteresting_key_lists = ([], [])
         for inv in self.repo.iter_inventories(inv_ids, 'unordered'):
             root_keys = root_keys_for_inv(inv)
             if (inv.revision_id,) in parent_invs_only_keys:
-                uninteresting.extend(root_keys)
+                key_lists = uninteresting_key_lists
             else:
-                interesting.extend(root_keys)
+                key_lists = interesting_key_lists
+            for key_list, key in zip(key_lists, root_keys):
+                key_list.append(key)
             present = no_fallback_chk_bytes_index.get_parent_map(root_keys)
             missing = set(root_keys).difference(present)
             all_missing.update([('chk_bytes',) + key for key in missing])
@@ -647,7 +650,24 @@
         chk_bytes_no_fallbacks._search_key_func = \
             self.repo.chk_bytes._search_key_func
         chk_diff = chk_map.iter_interesting_nodes(
-            chk_bytes_no_fallbacks, interesting, uninteresting)
+            chk_bytes_no_fallbacks, interesting_key_lists[0],
+            uninteresting_key_lists[0])
+        # XXX: this is basically the same as _filter_id_to_entry elsewhere in
+        # this file
+        bytes_to_info = inventory.CHKInventory._bytes_to_utf8name_key
+        text_keys = set()
+        try:
+            for interesting_rec, interesting_map in chk_diff:
+                for name, bytes in interesting_map:
+                    _, file_id, revision_id = bytes_to_info(bytes)
+                    text_keys.add((file_id, revision_id))
+        except errors.NoSuchRevision, e:
+            raise errors.BzrCheckError(
+                "Repository %s missing chk node(s) for new revisions."
+                % (self.repo,))
+        chk_diff = chk_map.iter_interesting_nodes(
+            chk_bytes_no_fallbacks, interesting_key_lists[1],
+            uninteresting_key_lists[1])
         try:
             for interesting_rec, interesting_map in chk_diff:
                 pass
@@ -655,7 +675,13 @@
             raise errors.BzrCheckError(
                 "Repository %s missing chk node(s) for new revisions."
                 % (self.repo,))
-        
+        present_text_keys = no_fallback_texts_index.get_parent_map(text_keys)
+        missing_text_keys = text_keys.difference(present_text_keys)
+        if missing_text_keys:
+            raise errors.BzrCheckError(
+                "Repository %s missing text keys %r for new revisions"
+                 % (self.repo, sorted(missing_text_keys)))
+
     def _execute_pack_operations(self, pack_operations,
                                  _packer_class=GCCHKPacker,
                                  reload_func=None):
@@ -1123,7 +1149,7 @@
             missing_inventory_keys.add(key[1:])
         if self._chk_id_roots or self._chk_p_id_roots:
             raise AssertionError('Cannot call get_stream_for_missing_keys'
-                ' untill all of get_stream() has been consumed.')
+                ' until all of get_stream() has been consumed.')
         # Yield the inventory stream, so we can find the chk stream
         # Some of the missing_keys will be missing because they are ghosts.
         # As such, we can ignore them. The Sink is required to verify there are

=== modified file 'bzrlib/tests/per_repository/test_write_group.py'
--- bzrlib/tests/per_repository/test_write_group.py	2009-09-07 05:48:51 +0000
+++ bzrlib/tests/per_repository/test_write_group.py	2009-09-07 07:41:17 +0000
@@ -518,16 +518,10 @@
         """commit_write_group fails with BzrCheckError when the chk root record
         for a parent inventory of a new revision is missing.
         """
-        builder = self.make_branch_builder('simple-branch')
-        builder.build_snapshot('A-id', None, [
-            ('add', ('', 'root-id', 'directory', None)),
-            ('add', ('file', 'file-id', 'file', 'content\n'))])
-        builder.build_snapshot('B-id', None, [])
-        builder.build_snapshot('C-id', None, [
-            ('modify', ('file-id', 'new-content'))])
-        b = builder.get_branch()
-        if not b.repository._format.supports_chks:
+        repo = self.make_repository('damaged-repo')
+        if not repo._format.supports_chks:
             raise TestNotApplicable('requires repository with chk_bytes')
+        b = self.make_branch_with_multiple_chk_nodes()
         b.lock_read()
         self.addCleanup(b.unlock)
         # Now, manually insert objects for a stacked repo with only revision
@@ -538,7 +532,6 @@
         inv_c = b.repository.get_inventory('C-id')
         chk_keys_for_c_only = [
             inv_c.id_to_entry.key(), inv_c.parent_id_basename_to_file_id.key()]
-        repo = self.make_repository('damaged-repo')
         repo.lock_write()
         repo.start_write_group()
         src_repo = b.repository
@@ -560,6 +553,65 @@
             errors.BzrCheckError, reopened_repo.commit_write_group)
         reopened_repo.abort_write_group()
 
+    def make_branch_with_multiple_chk_nodes(self):
+        # add and modify files with very long file-ids, so that the chk map
+        # will need more than just a root node.
+        builder = self.make_branch_builder('simple-branch')
+        file_adds = []
+        file_modifies = []
+        for char in 'abc':
+            name = char * 10000
+            file_adds.append(
+                ('add', ('file-' + name, 'file-%s-id' % name, 'file',
+                         'content %s\n' % name)))
+            file_modifies.append(
+                ('modify', ('file-%s-id' % name, 'new content %s\n' % name)))
+        builder.build_snapshot('A-id', None, [
+            ('add', ('', 'root-id', 'directory', None))] +
+            file_adds)
+        builder.build_snapshot('B-id', None, [])
+        builder.build_snapshot('C-id', None, file_modifies)
+        return builder.get_branch()
+        
+    def test_missing_text_record(self):
+        """commit_write_group fails with BzrCheckError when a text is missing.
+        """
+        repo = self.make_repository('damaged-repo')
+        if not repo._format.supports_chks:
+            raise TestNotApplicable('requires repository with chk_bytes')
+        b = self.make_branch_with_multiple_chk_nodes()
+        src_repo = b.repository
+        src_repo.lock_read()
+        self.addCleanup(src_repo.unlock)
+        # Now, manually insert objects for a stacked repo with only revision
+        # C-id, *except* drop one changed text (one that from a non-root chk
+        # page)
+        # Pick a non-root key to drop
+        all_texts = src_repo.texts.keys()
+        all_texts.remove(('file-%s-id' % ('c'*10000,), 'C-id'))
+        repo.lock_write()
+        repo.start_write_group()
+        repo.chk_bytes.insert_record_stream(
+            src_repo.chk_bytes.get_record_stream(
+                src_repo.chk_bytes.keys(), 'unordered', True))
+        repo.texts.insert_record_stream(
+            src_repo.texts.get_record_stream(
+                all_texts, 'unordered', True))
+        repo.inventories.insert_record_stream(
+            src_repo.inventories.get_record_stream(
+                [('B-id',), ('C-id',)], 'unordered', True))
+        repo.revisions.insert_record_stream(
+            src_repo.revisions.get_record_stream(
+                [('C-id',)], 'unordered', True))
+        # Make sure the presence of the missing data in a fallback does not
+        # avoid the error.
+        repo.add_fallback_repository(b.repository)
+        self.assertRaises(errors.BzrCheckError, repo.commit_write_group)
+        reopened_repo = self.reopen_repo_and_resume_write_group(repo)
+        self.assertRaises(
+            errors.BzrCheckError, reopened_repo.commit_write_group)
+        reopened_repo.abort_write_group()
+
 
 class TestResumeableWriteGroup(TestCaseWithRepository):
 

